@@ -373,6 +373,14 @@ window_decorated_notify (MetaWindow *mw,
   meta_window_actor_constructed (G_OBJECT (self));
 }
 
+static void
+window_appears_focused_notify (MetaWindow *mw,
+                               GParamSpec *arg1,
+                               gpointer    data)
+{
+  clutter_actor_queue_redraw (CLUTTER_ACTOR (data));
+}
+
 static void
 meta_window_actor_constructed (GObject *object)
 {
@@ -403,8 +411,6 @@ meta_window_actor_constructed (GObject *object)
   if (format && format->type == PictTypeDirect && format->direct.alphaMask)
     priv->argb32 = TRUE;
 
-  meta_window_actor_update_opacity (self);
-
   if (!priv->actor)
     {
       priv->actor = meta_shaped_texture_new ();
@@ -422,6 +428,8 @@ meta_window_actor_constructed (GObject *object)
 
       g_signal_connect (priv->window, "notify::decorated",
                         G_CALLBACK (window_decorated_notify), self);
+      g_signal_connect (priv->window, "notify::appears-focused",
+                        G_CALLBACK (window_appears_focused_notify), self);
     }
   else
     {
@@ -432,7 +440,7 @@ meta_window_actor_constructed (GObject *object)
       clutter_actor_raise_top (priv->actor);
     }
 
-
+  meta_window_actor_update_opacity (self);
   meta_window_actor_update_shape (self, priv->shaped);
 }
 
@@ -497,6 +505,12 @@ meta_window_actor_dispose (GObject *object)
 
   info->windows = g_list_remove (info->windows, (gconstpointer) self);
 
+  if (priv->window)
+    {
+      g_object_unref (priv->window);
+      priv->window = NULL;
+    }
+
   /*
    * Release the extra reference we took on the actor.
    */
@@ -529,7 +543,11 @@ meta_window_actor_set_property (GObject      *object,
   switch (prop_id)
     {
     case PROP_META_WINDOW:
-      priv->window = g_value_get_object (value);
+      {
+        if (priv->window)
+          g_object_unref (priv->window);
+        priv->window = g_value_dup_object (value);
+      }
       break;
     case PROP_META_SCREEN:
       priv->screen = g_value_get_pointer (value);
@@ -684,6 +702,26 @@ meta_window_actor_get_shadow_bounds (MetaWindowActor       *self,
 }
 #endif
 
+/* If we have an ARGB32 window that we decorate with a frame, it's
+ * probably something like a translucent terminal - something where
+ * the alpha channel represents transparency rather than a shape.  We
+ * don't want to show the shadow through the translucent areas since
+ * the shadow is wrong for translucent windows (it should be
+ * translucent itself and colored), and not only that, will /look/
+ * horribly wrong - a misplaced big black blob. As a hack, what we
+ * want to do is just draw the shadow as normal outside the frame, and
+ * inside the frame draw no shadow.  This is also not even close to
+ * the right result, but looks OK. We also apply this approach to
+ * windows set to be partially translucent with _NET_WM_WINDOW_OPACITY.
+ */
+static gboolean
+clip_shadow_under_window (MetaWindowActor *self)
+{
+  MetaWindowActorPrivate *priv = self->priv;
+
+  return (priv->argb32 || priv->opacity != 0xff) && priv->window->frame;
+}
+
 static void
 meta_window_actor_paint (ClutterActor *actor)
 {
@@ -696,17 +734,36 @@ meta_window_actor_paint (ClutterActor *actor)
     {
       MetaShadowParams params;
       cairo_rectangle_int_t shape_bounds;
+      cairo_region_t *clip = priv->shadow_clip;
 
       meta_window_actor_get_shape_bounds (self, &shape_bounds);
       meta_window_actor_get_shadow_params (self, appears_focused, &params);
 
+      /* The frame bounds are already subtracted from priv->shadow_clip
+       * if that exists.
+       */
+      if (!clip && clip_shadow_under_window (self))
+        {
+          cairo_region_t *frame_bounds = meta_window_get_frame_bounds (priv->window);
+          cairo_rectangle_int_t bounds;
+
+          meta_window_actor_get_shadow_bounds (self, appears_focused, &bounds);
+          clip = cairo_region_create_rectangle (&bounds);
+
+          cairo_region_subtract (clip, frame_bounds);
+        }
+
       meta_shadow_paint (shadow,
                          params.x_offset + shape_bounds.x,
                          params.y_offset + shape_bounds.y,
                          shape_bounds.width,
                          shape_bounds.height,
-                         (clutter_actor_get_paint_opacity (actor) * params.opacity) / 255,
-                         priv->shadow_clip);
+                         (clutter_actor_get_paint_opacity (actor) * params.opacity * priv->opacity) / (255 * 255),
+                         clip,
+                         clip_shadow_under_window (self)); /* clip_strictly - not just as an optimization */
+
+      if (clip && clip != priv->shadow_clip)
+        cairo_region_destroy (clip);
     }
 
   CLUTTER_ACTOR_CLASS (meta_window_actor_parent_class)->paint (actor);
@@ -1626,13 +1683,22 @@ meta_window_actor_update_bounding_region (MetaWindowActor *self,
 
   priv->bounding_region = cairo_region_create_rectangle (&bounding_rectangle);
 
-  /* When we're shaped, we use the shape region to generate the shadow; the shape
-   * region only changes when we get ShapeNotify event; but for unshaped windows
-   * we generate the shadow from the bounding region, so we need to recompute
-   * the shadow when the size changes.
-   */
-  if (!priv->shaped)
-    meta_window_actor_invalidate_shadow (self);
+  if (priv->shaped)
+    {
+      /* If we're shaped, the implicit shape region clipping we need to do needs
+       * to be updated.
+       */
+      meta_window_actor_update_shape (self, TRUE);
+    }
+  else
+    {
+      /* When we're shaped, we use the shape region to generate the shadow; the shape
+       * region only changes when we get ShapeNotify event; but for unshaped windows
+       * we generate the shadow from the bounding region, so we need to recompute
+       * the shadow when the size changes.
+       */
+      meta_window_actor_invalidate_shadow (self);
+    }
 
   g_signal_emit (self, signals[SIZE_CHANGED], 0);
 }
@@ -1653,6 +1719,20 @@ meta_window_actor_update_shape_region (MetaWindowActor *self,
       cairo_rectangle_int_t rect = { rects[i].x, rects[i].y, rects[i].width, rects[i].height };
       cairo_region_union_rectangle (priv->shape_region, &rect);
     }
+
+  /* Our "shape_region" is called the "bounding region" in the X Shape
+   * Extension Documentation.
+   *
+   * Our "bounding_region" is called the "bounding rectangle", which defines
+   * the shape of the window as if it the window was unshaped.
+   *
+   * The X Shape extension requires that the "bounding region" can never
+   * extend outside the "bounding rectangle", and says it must be implicitly
+   * clipped before rendering. The region we get back hasn't been clipped.
+   * We explicitly clip the region here.
+   */
+  if (priv->bounding_region != NULL)
+    cairo_region_intersect (priv->shape_region, priv->bounding_region);
 }
 
 /**
@@ -1670,7 +1750,7 @@ meta_window_actor_get_obscured_region (MetaWindowActor *self)
 {
   MetaWindowActorPrivate *priv = self->priv;
 
-  if (!priv->argb32 && priv->back_pixmap)
+  if (!priv->argb32 && priv->opacity == 0xff && priv->back_pixmap)
     {
       if (priv->shaped)
         return priv->shape_region;
@@ -1769,6 +1849,12 @@ meta_window_actor_set_visible_region_beneath (MetaWindowActor *self,
     {
       meta_window_actor_clear_shadow_clip (self);
       priv->shadow_clip = cairo_region_copy (beneath_region);
+
+      if (clip_shadow_under_window (self))
+        {
+          cairo_region_t *frame_bounds = meta_window_get_frame_bounds (priv->window);
+          cairo_region_subtract (priv->shadow_clip, frame_bounds);
+        }
     }
 }
 
@@ -2127,5 +2213,5 @@ meta_window_actor_update_opacity (MetaWindowActor *self)
     opacity = 255;
 
   self->priv->opacity = opacity;
-  clutter_actor_set_opacity (CLUTTER_ACTOR (self), opacity);
+  clutter_actor_set_opacity (self->priv->actor, opacity);
 }
