@@ -107,9 +107,6 @@ meta_core_get (Display *xdisplay,
       case META_CORE_GET_CLIENT_HEIGHT:
         *((gint*)answer) = window->rect.height;
         break;
-      case META_CORE_IS_TITLEBAR_ONSCREEN:
-        *((gboolean*)answer) = meta_window_titlebar_is_onscreen (window);
-        break;
       case META_CORE_GET_CLIENT_XWINDOW:
         *((Window*)answer) = window->xwindow;
         break;
@@ -146,6 +143,9 @@ meta_core_get (Display *xdisplay,
       case META_CORE_GET_FRAME_HEIGHT:
         *((gint*)answer) = window->frame->rect.height;
         break;
+      case META_CORE_GET_THEME_VARIANT:
+        *((char**)answer) = window->gtk_theme_variant;
+        break;
       case META_CORE_GET_SCREEN_WIDTH:
         *((gint*)answer) = window->screen->rect.width;
         break;
